const { commas, saveGame, quackGradeMultiplier, prestigeMultiplier, makeBackup } = require('./utils')
const { quackStore } = require('./quackstore')

const possiblePrestige = coins => {
  let p = 0
  while (totalCostForPrestige(p + 1) <= coins) {
    p += 1
  }
  return p
}

const tpcRecMemo = []
const totalCostForPrestige = prestigeLevel => {
  if (prestigeLevel === 0) {
    return 0
  }
  return (tpcRecMemo[prestigeLevel]) || (tpcRecMemo[prestigeLevel] = 1_000_000_000_000 * Math.pow(prestigeLevel, 3) + totalCostForPrestige(prestigeLevel - 1))
}

const prestigeRoute = async ({ say, words, user }) => {
  const possible = possiblePrestige(user.coinsAllTime)
  const current = user.prestige ??= 0
  if (words[1] === 'me') {
    await say(
      'This will permanently remove all of your items, upgrades, and coins!\n\n' +
            'Say \'!!prestige me\' to confirm.'
    )
  } else {
    await say(
      `Current Prestige: ${commas(current)}\n\n` +
      `Quacks gained if you prestige now: ${commas(possible - current)}\n\n` +
      `HVAC until next quack: ${commas(totalCostForPrestige(possible + 1) - user.coinsAllTime)}\n\n` +
      'Say \'!prestige me\' to start the prestige process.' +
      `\n\nYour prestige is currently boosting your CPS by ${commas((prestigeMultiplier(user) - 1) * 100)}%`
    )
  }
}//, true, adminOnly)

const prestigeConfirmRoute = async ({ event, say, user }) => {
  const possible = possiblePrestige(user.coinsAllTime)
  const current = user.prestige
  if (possible <= current) {
    await say('You don\'t have enough HVAC to prestige right now!')
    return
  }
  if (event?.text !== '!!prestige me') {
    await say('Say exactly \'!!prestige me\' to confirm')
    return
  }
  await makeBackup()

  user.quacks ??= 0
  user.quacks += (possible - user.prestige)

  user.prestige = possible
  user.coins = 0
  user.items = {}
  user.upgrades = {}

  saveGame()
  await say('You prestiged! Check out !quackstore to see what you can buy!')
}

const quackStoreListing = (showCost = true) => ([name, upgrade]) =>
    showCost
      ? `:${upgrade.emoji}: *${name}* - Costs *${upgrade.cost} Quack.*\n\n_${upgrade.description}_`
      : `:${upgrade.emoji}: *${name}* - Worth *${upgrade.cost} Quack.*\n\n_${upgrade.description}_`

const allUserQuackUpgrades = user =>
  Object.entries(user.quackUpgrades || {})
    .map(([type, upgrades]) => upgrades).flatMap(x => x)

const hasPreReqs = user => ([name, upgrade]) => {
  if (!upgrade.preReqs) {
    return true
  }
  const allUserUpgrades = allUserQuackUpgrades(user)
  return upgrade.preReqs.every(preReq => allUserUpgrades.includes(preReq))
}

const owns = (user, [name, upgrade]) => allUserQuackUpgrades(user).includes(name)

const ownedQuackItems = user => Object.entries(quackStore).filter(upgrade => owns(user, upgrade))

const unownedQuackItems = user => Object.entries(quackStore).filter(upgrade => !owns(user, upgrade))

const quackStoreText = user =>
  unownedQuackItems(user)
    .filter(hasPreReqs(user))
    .map(quackStoreListing(true))
    .join('\n\n') +
  `\n\nYou have ${user.quacks ??= 0} quacks to spend.` +
  `\nQuackStore upgrades are currently boosting your CPS by ${commas((quackGradeMultiplier(user) - 1) * 100)}%`

const quackStoreRoute = async ({ user, say, words }) => {
  user.quackUpgrades ??= {}
  const quacks = user.quacks ??= 0
  if (!words[1]) {
    await say(quackStoreText(user))
    return
  }
  console.log(`Trying to buy ${words[1]}`)
  const quackItem = quackStore[words[1]]
  if (!quackItem || !unownedQuackItems(user).find(([name]) => name === words[1])) {
    await say(`'${words[1]}' is not available in the quack store!`)
    return
  }
  if (quackItem.cost > quacks) {
    await say(`${words[1]} costs ${quackItem.cost} Quacks, but you only have ${quacks}!`)
    return
  }
  user.quacks -= quackItem.cost
  user.quackUpgrades[quackItem.type] ??= []
  user.quackUpgrades[quackItem.type].push(words[1])
  saveGame()
}

const ownedQuacksText = user =>
  ownedQuackItems(user)
    .filter(hasPreReqs(user))
    .map(quackStoreListing(false))
    .join('\n\n') +
  `\n\nQuackStore upgrades are currently boosting your CPS by ${commas((quackGradeMultiplier(user) - 1) * 100)}%`

const ownedQuacksRoute = async ({ say, user }) => {
  user.quackUpgrades ??= {}
  user.quacks ??= 0
  await say(ownedQuacksText(user))
}

module.exports = {
  quackStoreRoute,
  prestigeRoute,
  prestigeConfirmRoute,
  ownedQuacksRoute
}
