const basic = ({ type, description, count, cost, extraCondition = () => true, effect = cps => cps * 2 }) => ({
  type,
  description,
  condition: (user, squadGrades) => user.items[type] >= count && extraCondition(user, squadGrades),
  cost,
  effect
})

const evil = ({ type, description, cost }) => basic({
  type,
  description,
  count: 40,
  cost,
  extraCondition: (user, squadGrades) => squadGrades?.includes('discardHumanMorals'),
})

const heavenly = ({ type, description, cost }) => basic({
  type,
  description,
  count: 60,
  cost,
  extraCondition: (user, squadGrades) => squadGrades?.includes('redemption'),
})

const baby = ({ type, description, cost }) => basic({
  type,
  description,
  count: 80,
  cost,
  extraCondition: (user, squadGrades) => squadGrades?.includes('redemption'),
})

const geometry = ({ type, description, cost }) => basic({
  type,
  description,
  count: 100,
  cost,
  extraCondition: (user, squadGrades) => squadGrades?.includes('redemption'),
})

const universitality = ({ type, description, cost }) => basic({
  type,
  description,
  count: 100,
  cost,
  extraCondition: (user, squadGrades) => squadGrades?.includes('redemption'),
})

module.exports = {
  doubleClick: basic({
    type: 'mouse',
    description: 'Doubles the power of mice',
    count: 1,
    cost: 1_000
  }),
  stinkierCheese: basic({
    type: 'mouse',
    description: 'Mice are doubly motivated to hunt down HVAC Coins',
    count: 10,
    cost: 21_000
  }),
  biggerTeeth: basic({
    type: 'mouse',
    description: 'Mice can intimidate twice as much HVAC out of their victims.',
    count: 25,
    cost: 50_000
  }),
  rats: evil({
    type: 'mouse',
    description: 'Consume the rotten remains of your foes',
    cost: 150_000,
  }),
  hoodedMice: heavenly({
    type: 'mouse',
    description: 'These monks have nearly reached enlightenment.',
    cost: 1_000_000,
  }),

  fasterComputers: basic({
    type: 'accountant',
    description: 'Accountants can ~steal~ optimize twice as much HVAC!',
    count: 1,
    cost: 11_000,
  }),
  lackOfMorality: basic({
    type: 'accountant',
    description: 'Accountants are taking a hint from nearby CEOs.',
    count: 10,
    cost: 200_000,
  }),
  widerBrains: basic({
    type: 'accountant',
    description: 'For accountant do double of thinking.',
    count: 25,
    cost: 550_000,
  }),
  vastLayoffs: evil({
    type: 'accountant',
    description: 'The weak are not part of our future.',
    cost: 2_450_000,
  }),
  charityFund: heavenly({
    type: 'accountant',
    description: 'THIS one is more than just a tax break.',
    cost: 16_333_333,
  }),

  biggerBlowhole: basic({
    type: 'whale',
    description: 'With all that extra air, whales have double power!',
    count: 1,
    cost: 120_000
  }),
  sassyWhales: basic({
    type: 'whale',
    description: 'These are the kind of whales that know how to get twice as much done',
    count: 10,
    cost: 3_000_000
  }),
  thinnerWater: basic({
    type: 'whale',
    description: 'Whales can move twice as quickly through this physics-defying liquid',
    count: 25,
    cost: 6_000_000
  }),
  blightWhales: evil({
    type: 'whale',
    description: `Infectious with evil, they swim the ocean spreading their spores.`,
    cost: 24_000_000
  }),
  whaleChoir: heavenly({
    type: 'whale',
    description: `Their cleansing songs reverberate through the sea.`,
    cost: 144_000_000
  }),

  greasyTracks: basic({
    type: 'train',
    description: 'Lets trains deliver HVAC twice as efficiently',
    count: 1,
    cost: 1_300_000
  }),
  rocketThrusters: basic({
    type: 'train',
    description: 'That\'ll put some quack on your track',
    count: 10,
    cost: 22_000_000
  }),
  loudConductors: basic({
    type: 'train',
    description: 'Conductors can onboard twice as much HVAC',
    count: 25,
    cost: 65_000_000
  }),
  hellTrain: evil({
    type: 'train',
    description: 'Shipping blood needed for the ritual.',
    cost: 370_000_000
  }),
  toyTrain: heavenly({
    type: 'train',
    description: 'Something simple. Toot toot!',
    cost: 2_220_000_000
  }),

  gasolineFire: basic({
    type: 'fire',
    description: 'Extremely good for breathing in.',
    count: 1,
    cost: 14_000_000
  }),
  extremelyDryFuel: basic({
    type: 'fire',
    description: 'Use the ignite command for a secret achievement.',
    count: 10,
    cost: 163_000_000
  }),
  cavemanFire: basic({
    type: 'fire',
    description: 'They just don\'t make \'em like they used to.',
    count: 25,
    cost: 700_000_000
  }),
  lava: evil({
    type: 'fire',
    description: `Hopefully no usurpers have any "accidents".`,
    cost: 4_200_000_000
  }),
  blueFire: heavenly({
    type: 'fire',
    description: `You can hear it singing with delight.`,
    cost: 25_200_000_000
  }),

  spoonerang: basic({
    type: 'boomerang',
    description: 'Scoops up HVAC mid-flight',
    count: 1,
    cost: 200_000_000
  }),
  boomerAng: basic({
    type: 'boomerang',
    description: 'It\'s... old.',
    count: 10,
    cost: 1_200_000_000
  }),
  doubleRang: basic({
    type: 'boomerang',
    description: 'You throw one, but somehow catch two',
    count: 25,
    cost: 10_000_000_000
  }),
  loyalRang: evil({
    type: 'boomerang',
    description: `Frequently reports back to your throne on the state of your empire.`,
    cost: 60_000_000_000
  }),
  youRang: heavenly({
    type: 'boomerang',
    description: 'Your arms and legs recede into your body. You bend at the middle. You fly. And for a moment, you are free.',
    cost: 360_000_000_000
  }),

  lunarPower: basic({
    type: 'moon',
    description: 'Out with the sol, in with the lun!',
    count: 1,
    cost: 3_300_000_000
  }),
  womanOnTheMoon: basic({
    type: 'moon',
    description: 'There\'s no reason for it not to be a woman!',
    count: 10,
    cost: 39_700_000_000
  }),
  doubleCraters: basic({
    type: 'moon',
    description: 'Making every side look like the dark side.',
    count: 25,
    cost: 165_000_000_000
  }),
  tidalUpheaval: evil({
    type: 'moon',
    description: `The hell with the ocean. That's valuable_ *the abstract concept of more power* _we're losing.`,
    cost: 865_000_000_000
  }),
  newMoon: heavenly({
    type: 'moon',
    description: `Build a second moon to provide space for affordable housing.`,
    cost: 5_190_000_000_000
  }),

  glassButterfly: basic({
    type: 'butterfly',
    description: 'Not your grandma\'s universe manipulation.',
    count: 1,
    cost: 51_000_000_000
  }),
  monarchMigration: basic({
    type: 'butterfly',
    description: 'This upgrade brought to you by milkweed.',
    count: 10,
    cost: 870_000_000_000
  }),
  quadWing: basic({
    type: 'butterfly',
    description: 'Sounds a lot like a trillion bees buzzing inside your head.',
    count: 25,
    cost: 2_550_000_000_000
  }),
  venomousMoths: evil({
    type: 'butterfly',
    description: 'Specifically manufactured for their horrifying brain-melt toxins.',
    cost: 12_550_000_000_000
  }),
  quietingNectar: heavenly({
    type: 'butterfly',
    description: 'Calming and extra sweet. Soothes even human ails.',
    cost: 75_300_000_000_000
  }),

  silverMirror: basic({
    type: 'mirror',
    description: 'Excellent for stabbing vampires.',
    count: 1,
    cost: 750_000_000_000
  }),
  pocketMirror: basic({
    type: 'mirror',
    description: 'Take your self-reflection on the go!',
    count: 10,
    cost: 18_000_000_000_000
  }),
  window: basic({
    type: 'mirror',
    description: 'Only through looking around you can you acquire the self reflection necessary to control the thermostat.',
    count: 25,
    cost: 37_500_000_000_000
  }),
  crackedMirror: evil({
    type: 'mirror',
    description: `YOU SMILE. DO NOT FEAR, THIS IS THE FACE OF A FRIEND.`,
    cost: 222_000_000_000_000
  }),
  funHouseMirror: heavenly({
    type: 'mirror',
    description: `yoU LOok so siLLY IN thesE THINgs`,
    cost: 1_330_000_000_000_000
  }),

  fzero: basic({
    type: 'quade',
    description: 'Brings out his competitive spirit.',
    count: 1,
    cost: 10_000_000_000_000
  }),
  triHumpCamel: basic({
    type: 'quade',
    description: 'YEE HAW :trimedary_camel:',
    count: 10,
    cost: 200_000_000_000_000
  }),
  adam: basic({
    type: 'quade',
    description: 'He could probably reach the thermostat if he wanted.',
    count: 25,
    cost: 500_000_000_000_000
  }),
  thatsNotQuade: evil({
    type: 'quade',
    description: `The skinless face lacks even a moustache. Nevertheless, it pledges its allegiance.`,
    cost: 3_000_000_000_000_000
  }),
  hannahMontanaLinux: heavenly({
    type: 'quade',
    description: `The patrician's choice.`,
    cost: 18_000_000_000_000_000
  }),

  latestNode: basic({
    type: 'hvacker',
    description: 'The old one has terrible ergonomics, tsk tsk.',
    count: 1,
    cost: 140_000_000_000_000
  }),
  nativeFunctions: basic({
    type: 'hvacker',
    description: 'Sometimes javascript just isn\'t fast enough.',
    count: 10,
    cost: 3_300_000_000_000_000
  }),
  gitCommits: basic({
    type: 'hvacker',
    description: 'The heads of multiple people in a company are better than, for example, merely one head.',
    count: 25,
    cost: 7_000_000_000_000_000
  }),
  undefinedBehavior: evil({
    type: 'hvacker',
    description: `skREEEFDS☐☐☐☐☐it's☐jwtoo☐laate☐☐☐☐☐`,
    cost: 42_000_000_000_000_000
  }),
  mutualUnderstanding: heavenly({
    type: 'hvacker',
    description: `lol fat chance, dummy. Points for trying, though`,
    cost: 250_000_000_000_000_000
  }),

  coffee: basic({
    type: 'creator',
    description: `Didn't you know? It makes you smarter. No consequencAAAAAA`,
    count: 1,
    cost: 1_960_000_000_000_000
  }),
  bribery: basic({
    type: 'creator',
    description: `How much could he be making that a couple bucks won't get me more HVAC?`,
    count: 10,
    cost: 32_300_000_000_000_000
  }),
  vim: basic({
    type: 'creator',
    description: `*teleports behind you*`,
    count: 25,
    cost: 100_000_000_000_000_000
  }),
  regrets: evil({
    type: 'creator',
    description: `HE HAS NONE. HE LAUGHS.`,
    cost: 600_000_000_000_000_000
  }),
  goVegan: heavenly({
    type: 'creator',
    description: `Unlock your vegan powers.`,
    cost: 3_600_000_000_000_000_000
  }),

  angelInvestors: basic({
    type: 'smallBusiness',
    description: 'Not so small NOW are we?',
    count: 1,
    cost: 3_140_000_000_000_000
  }),
  officeManager: basic({
    type: 'smallBusiness',
    description: 'Sate your laborers with snacks.',
    count: 10,
    cost: 80_000_000_000_000_000
  }),
  undyingLoyalty: basic({
    type: 'smallBusiness',
    description: 'Your foolish employees bow to your every whim, regardless of salary.',
    count: 25,
    cost: 138_000_000_000_000_000
  }),
  deathSquad: evil({
    type: 'smallBusiness',
    description: `pwease don't unionize uwu :pleading_face:`,
    cost: 858_000_000_000_000_000
  }),
  coop: heavenly({
    type: 'smallBusiness',
    description: `By the people, for the people.`,
    cost: 5_140_000_000_000_000_000
  }),

  corporateBuyouts: basic({
    type: 'bigBusiness',
    description: 'The cornerstone of any family-run business.',
    count: 1,
    cost: 28_140_000_000_000_000
  }),
  politicalSway: basic({
    type: 'bigBusiness',
    description: `What's a bit of lobbying between friends?`,
    count: 10,
    cost: 560_000_000_000_000_000
  }),
  humanDiscontent: basic({
    type: 'bigBusiness',
    description: 'A sad populace is a spendy populace!',
    count: 25,
    cost: 1_372_000_000_000_000_000
  }),
  weJustKillPeopleNow: evil({
    type: 'bigBusiness',
    description: 'It is extremely difficult to get more evil than we already were. Nevertheless,',
    cost: 7_072_000_000_000_000_000
  }),
  makePublic: heavenly({
    type: 'bigBusiness',
    description: `Downplay immediate profit for more long-term benefits.`,
    cost: 42_000_000_000_000_000_000
  }),

  homage: {
    type: 'general',
    description: 'The power of original ideas increases your overall CPS by 10%',
    condition: user => Object.entries(user.items).reduce((total, [, countOwned]) => countOwned + total, 0) >= 200,
    emoji: 'cookie',
    cost: 10_000_000_000,
    effect: (itemCps, user) => itemCps * 1.1
  },
  iLoveHvac: {
    type: 'general',
    description: 'The power of love increases your overall CPS by 10%',
    condition: user => Object.entries(user.items).reduce((total, [, countOwned]) => countOwned + total, 0) >= 400,
    emoji: 'heart',
    cost: 100_000_000_000_000,
    effect: (itemCps, user) => itemCps * 1.1
  }
}
